/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Enrique Robledo Arnuncio, for digitalSTROM AG <era@tegulan.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __VDSM_COORDINATION_H__
#define __VDSM_COORDINATION_H__

#include "role.h"

/**
 * @file coordination.h
 *
 * @brief Utilities used by discovery.c to coordinate with other vdsm instances.
 *
 * All the functions in this header should be called exclusively from the avahi
 * service thread, except for vdsm_coordination_is_collection_enabled(), which
 * is thread safe, and vdsm_coordination_set_collection_initialized_callback(),
 * which should be called before the avahi service thread starts.
 */

typedef void (*collection_initialized_callback_t)(bool collection_enabled);

/**
 * @brief Sets the coordination post-initialization callback.
 *
 * The provided callback will be called exactly once after the first avahi vdsm
 * browser sweep has completed. The collection_enabled argument passed to the
 * callback will be false if there was some foreign vdsm discovered during
 * initialization, even if it has since been removed. collection_enabled will be
 * true only if there were no other vdsms detected during initialization.
 *
 * The callback will be called from the avahi service thread, triggered by the
 * call to vdsm_coordination_initial_browse_complete().
 *
 * @param callback The callback function.
 */
void vdsm_coordination_set_collection_initialized_callback(
    collection_initialized_callback_t callback);

/**
 * @brief Getter for the collection enabled state.
 *
 * @return false if some other vdsm is present in the network, or during
 * initialization, until the absence or other vdsms can be confirmed.
 */
bool vdsm_coordination_is_collection_enabled();

/**
 * @brief Called to register a foreign vdsm present in the network.
 *
 * This function should be called as soon as the avahi vdsm browser discovers
 * a new vdsm.
 *
 * @param service_name The service name in the foreign vdsm avahi announcement.
 */
void vdsm_coordination_add_foreign_vdsm(const char* service_name);

/**
 * @brief Called to set the role of a foreign vdsm present in the network.
 *
 * @param service_name The service name in the foreign vdsm avahi announcement.
 * @param role The vdsm role information from the avahi resolve response.
 */
void vdsm_coordination_set_foreign_vdsm_role(
    const char* service_name,
    const VDSMRole* role);

/**
 * @brief Called to indicate the removal of a foreign vdsm from the network.
 *
 * @param service_name The service name in the foreign vdsm avahi announcement.
 */
void vdsm_coordination_remove_foreign_vdsm(const char* service_name);

/**
 * @brief Called to indicate the completion of the initial avahi vdsm browse
 *
 * This function should be called exactly once. I will trigger a call to the
 * callback function specified with
 * vdsm_coordination_set_collection_initialized_callback
 */
void vdsm_coordination_initial_browse_complete();

#endif  //__VDSM_COORDINATION_H__
