/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "client.h"
#include <boost/foreach.hpp>
#include <avahi-common/error.h>
#include <ds/str.h>
#include "entry-group.h"
#include "util.h"

typedef boost::system::error_code error_code;

namespace ds {
namespace avahi {

Client::Client(Poll &poll) : m_poll(poll), m_active(false) {
    VDSM_ENTER();
    reconnect();
}

void Client::addEntryGroup(EntryGroup &entryGroup) {
    m_entryGroups.push_back(&entryGroup);
    try {
        entryGroup.setActive(m_active);
    } catch (std::exception &e) {
        handleError(e);
    }
}

void Client::throwError(const char *message) {
    throwError(message, avahi_client_errno(get()));
}

void Client::throwError(const char *message, int error) {
    const char* strerror = avahi_strerror(error);
    throw std::runtime_error(ds::str(message, " error:", error, '-', strerror));
}

void Client::handleError(const std::exception &e) {
    VDSM_ERROR("%s e.what():%s", __PRETTY_FUNCTION__, e.what());
    reconnect();
}

void Client::setActive(bool active) {
    if (m_active == active) {
        return;
    }
    m_active = active;
    BOOST_FOREACH(auto&& entryGroup, m_entryGroups) {
        entryGroup->setActive(m_active);
    }
}

void Client::reconnect() {
    VDSM_ENTER();
    setActive(false);
    m_ptr.reset();

    // TODO(someday): exponential backoff
    int error;
    m_ptr.reset(avahi_client_new(m_poll.get(), AVAHI_CLIENT_NO_FAIL, sCallback, this, &error));
    if (!m_ptr) {
        // According to documentation, this should not happen
        // because we pass AVAHI_CLIENT_NO_FAIL flag.

        // TODO(someday): reconnect. But it needs to be asynchronous on timer to avoid busy loop
        handleError(std::runtime_error(ds::str("Failed to create client. error:", avahi_strerror(error))));
    }
}

void Client::sCallback(AvahiClient *, AvahiClientState state, void *data) {
    auto&& x = *static_cast<Client*>(data);
    try {
        x.callback(state);
    } catch (const std::exception &e) {
        x.handleError(e);
    }
}

void Client::callback(AvahiClientState state) {
    VDSM_ENTER("state:%d", int(state));
    switch (state) {
        case AVAHI_CLIENT_S_RUNNING:
            // The server has startup successfully and registered its host
            // name on the network, so it's time to create our services
            setActive(true);
            break;

        case AVAHI_CLIENT_S_COLLISION:
            // Let's drop our registered services. When the server is back
            // in AVAHI_SERVER_RUNNING state we will register them
            // again with the new host name.
        case AVAHI_CLIENT_S_REGISTERING:
            // The server records are now being established. This
            // might be caused by a host name change. We need to wait
            // for our own records to register until the host name is
            // properly established.
            setActive(false);
            break;
        case AVAHI_CLIENT_FAILURE:
            throwError("Avahi client failure");
            break;
        case AVAHI_CLIENT_CONNECTING:
            // Nothing to do. We wait for avahi deamon to be started.
            break;
    }
}

} // namespace avahi
} // namespace ds
