/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "entry-group.h"
#include <boost/foreach.hpp>
#include <avahi-common/alternative.h>
#include <avahi-common/error.h>
#include <avahi-common/malloc.h>
#include <ds/log.h>
#include <ds/str.h>
#include "util.h"
#include "client.h"

namespace ds {
namespace avahi {

struct AvahiCharPtrDeleter {
    void operator()(char* x) { avahi_free(x); }
};
typedef std::unique_ptr<char, AvahiCharPtrDeleter> UniqueAvahiCharPtr;

struct AvahiStringListDeleter {
    void operator()(AvahiStringList* x) { avahi_string_list_free(x); }
};
typedef std::unique_ptr<AvahiStringList, AvahiStringListDeleter> UniqueAvahiStringList;

EntryGroup::EntryGroup(Client& client, CollisionMode collisionMode, std::vector<Service>&& services)
    : m_client(client), m_collisionMode(collisionMode), m_services(std::move(services)) {}

void EntryGroup::publish() {
    if (!getActive()) {
        return;
    }

    avahi_entry_group_reset(get());

    if (m_services.empty()) {
        // avahi_entry_group_commit() must NOT be called for empty list of services
        return;
    }

    //TODO(now): asynchronous call with exponentional backoff to avoid inifinite cycles.

    BOOST_FOREACH(const auto& service, m_services) {
        UniqueAvahiStringList txtsList;
        auto&& txts = service.txts;
        if (!txts.empty()) {
            auto it = txts.begin();
            txtsList.reset(avahi_string_list_new(it->c_str(), NULL));
            it++;
            for (; it != txts.end(); ++it) {
                txtsList.reset(avahi_string_list_add(txtsList.release(), it->c_str()));
                DS_CHECK(txtsList);
            }
        }

        if (auto&& error = avahi_entry_group_add_service_strlst(get(), service.interface, service.protocol, service.flags,
            service.name.c_str(), service.type.c_str(), service.domain.c_str(), service.host.c_str(), service.port,
            txtsList.get())) {
          if (error == AVAHI_ERR_COLLISION) {
              handleCollision();
              return;
          }
          m_client.throwError("avahi_entry_group_add_service_strlst failed", error);
          return;
        }
    }
    if (auto&& error = avahi_entry_group_commit(get())) {
        m_client.throwError("avahi_entry_group_commit failed", error);
        return;
    }
}

void EntryGroup::setActive(bool active) {
    if (getActive() == active) {
        return;
    }
    m_ptr.reset();
    if (active) {
        m_ptr.reset(avahi_entry_group_new(m_client.get(), sCallback, this));
        if (!m_ptr) {
            m_client.throwError("Failed to create avahi entry group");
        }
        publish();
    }
}

void EntryGroup::handleCollision() {
    VDSM_ENTER("");
    switch (m_collisionMode) {
      case CollisionMode::ALTERNATIVE_NAME: {
        // Let's pick a new name.
        BOOST_FOREACH(auto&& service, m_services) {
            UniqueAvahiCharPtr name(avahi_alternative_service_name(service.name.c_str()));
            service.name = name.get();
            VDSM_NOTICE("Avahi service collision. renaming service to name:%s", service.name.c_str());
        }
        publish();
        break;
      }
    }
}

void EntryGroup::sCallback(AvahiEntryGroup *g, AvahiEntryGroupState state, void *data) {
    auto&& x = *static_cast<EntryGroup*>(data);
    try {
        x.callback(state);
    } catch (const std::exception &e) {
        x.m_client.handleError(e);
    }
}

void EntryGroup::callback(AvahiEntryGroupState state) {
    VDSM_ENTER("state:%d", int(state));
    switch (state) {
        case AVAHI_ENTRY_GROUP_ESTABLISHED:
            // The entries have successfully been established.
            BOOST_FOREACH(auto&& service, m_services) {
                VDSM_NOTICE("Avahi service established. name:%s", service.name.c_str());
            }
            break;
        case AVAHI_ENTRY_GROUP_COLLISION:
            // A name collision for one of the entries in the group has been detected,
            // the entries have been withdrawn.
            handleCollision();
            break;
        case AVAHI_ENTRY_GROUP_FAILURE:
            m_client.throwError("Avahi entry group failure");
            break;

        case AVAHI_ENTRY_GROUP_UNCOMMITED:
        case AVAHI_ENTRY_GROUP_REGISTERING:
            ;
    }
}

} // namespace avahi
} // namespace ds
