/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ReconnectingSocket.h"

#include <boost/asio/read.hpp>
#include <ds/asio/test.h>

typedef boost::system::error_code error_code;
typedef boost::asio::ip::tcp::resolver resolver;

namespace ds {
namespace vdsm {

class ReconnectingSocketTest : public ::testing::Test
{
public:
    boost::asio::io_service m_ioService;
    boost::asio::ip::tcp::acceptor m_acceptor;
    boost::asio::ip::tcp::socket m_acceptorSocket;
    //unique_ptr to support early delete
    std::unique_ptr<ReconnectingSocket> m_reconnectingSocket;
    std::array<char, 10> m_readArray;
    bool m_connectedCalled;
    bool m_disconnectedCalled;

    ReconnectingSocketTest()
        : m_ioService(1)
        , m_acceptor(m_ioService)
        , m_acceptorSocket(m_ioService)
        , m_reconnectingSocket(new ReconnectingSocket(m_ioService))
        , m_connectedCalled(false)
        , m_disconnectedCalled(false)
    {
        m_reconnectingSocket->setConnectTimeout(ds::asio::test::TIMEOUT);
        m_reconnectingSocket->setIdleTimeout(ds::asio::test::TIMEOUT);
        m_reconnectingSocket->connected().connect([=]()
        {
            m_connectedCalled = true;

            //start read to detect errors and start recovery
            boost::asio::async_read(m_reconnectingSocket->socket(), boost::asio::buffer(m_readArray),
                [=](const error_code& e, std::size_t n)
            {
                if (e == boost::asio::error::operation_aborted) {
                    return;
                }
                if (e) {
                    m_reconnectingSocket->reset();
                    return;
                }
                FAIL() << "We must not read anything";
            });
        });
        m_reconnectingSocket->disconnected().connect([=]()
        {
            m_disconnectedCalled = true;
        });
    }


    void acceptorAsyncListen(const boost::asio::ip::tcp::endpoint& endpoint)
    {
        m_acceptor.open(endpoint.protocol());
        m_acceptor.bind(endpoint);
        m_acceptor.listen(1); //queue size for listening connections
    }

    void acceptorAsyncListenV4(int port = 0)
    {
        acceptorAsyncListen(boost::asio::ip::tcp::endpoint(boost::asio::ip::tcp::v4(), port));
    }

    void acceptorAsyncListenV6(int port = 0)
    {
        acceptorAsyncListen(boost::asio::ip::tcp::endpoint(boost::asio::ip::tcp::v6(), port));
    }

    //Acceptor must be listening because we use its port
    void reconnectingSocketStart(int port)
    {
        EXPECT_FALSE(m_reconnectingSocket->isActive());
        m_reconnectingSocket->setQuery(resolver::query(
            "localhost",
            std::to_string(port)
        ));
        m_reconnectingSocket->start();
        EXPECT_TRUE(m_reconnectingSocket->isActive());
    }

    enum class Recovery {
        immediate,
        timeout
    };
    void runExpectConnected(Recovery recovery = Recovery::immediate)
    {
        ASSERT_FALSE(m_connectedCalled);
        ASSERT_FALSE(m_disconnectedCalled);
        ASSERT_FALSE(m_reconnectingSocket->socket().is_open());

        //We succeed when acceptor accepts the socket and client emits connected signal
        bool acceptedCalled = false;
        auto stopIfDone = [&]()
        {
            if (acceptedCalled && m_connectedCalled) {
                m_ioService.stop();
            }
        };
        m_acceptorSocket.close();
        m_acceptor.async_accept(m_acceptorSocket, [&](const error_code& e)
        {
            if (e == boost::asio::error::operation_aborted) {
                return;
            }
            if (e) {
                FAIL() << e.message();
            }
            acceptedCalled = true;
            stopIfDone();
        });

        if (recovery == Recovery::timeout) {
            //ensure recovery is not immediate
            ASSERT_RUN_NO_STOP(m_ioService);
        }

        ASSERT_FALSE(m_connectedCalled);
        ASSERT_FALSE(acceptedCalled);
        //run until until for connection confirmed on both sides
        boost::signals2::scoped_connection sc(
            m_reconnectingSocket->connected().connect([&]()
            {
                stopIfDone();
            })
        );
        ASSERT_TRUE(ds::asio::test::runStop(
            m_ioService,
            recovery == Recovery::immediate
                ? ds::asio::test::LATENCY / 2
                : ds::asio::test::LATENCY * 2
        ));

        EXPECT_TRUE(acceptedCalled);
        EXPECT_TRUE(m_acceptorSocket.is_open());
        EXPECT_TRUE(m_reconnectingSocket->socket().is_open());
        EXPECT_TRUE(m_connectedCalled);
        EXPECT_FALSE(m_disconnectedCalled);
        EXPECT_EQ(ReconnectingSocket::State::connected, m_reconnectingSocket->state());
        m_connectedCalled = false;
    }

    void runExpectDisconnected()
    {
        ASSERT_FALSE(m_connectedCalled);
        ASSERT_FALSE(m_disconnectedCalled);
        ASSERT_TRUE(m_acceptorSocket.is_open());
        ASSERT_TRUE(m_reconnectingSocket->socket().is_open());

        {
            boost::signals2::scoped_connection sc(m_reconnectingSocket->disconnected().connect([&]()
            {
                m_ioService.stop();
            }));
            ASSERT_RUN_STOP(m_ioService);
        }

        EXPECT_FALSE(m_reconnectingSocket->socket().is_open());
        EXPECT_FALSE(m_connectedCalled);
        EXPECT_TRUE(m_disconnectedCalled);
        EXPECT_EQ(ReconnectingSocket::State::disconnected, m_reconnectingSocket->state());

        m_disconnectedCalled = false;
        //TODO wait for closed m_acceptorSocket (we must read something from socket)

    }

    void runExpectReconnected()
    {
        m_reconnectingSocket->setIdleTimeout(ds::asio::test::LATENCY);
        EXPECT_EQ(ds::asio::test::LATENCY, m_reconnectingSocket->idleTimeout());
        runExpectDisconnected();
        runExpectConnected(Recovery::timeout);
        m_reconnectingSocket->setIdleTimeout(ds::asio::test::TIMEOUT);
    }
};

TEST_F(ReconnectingSocketTest, connectV4)
{
    acceptorAsyncListenV4();
    auto port = m_acceptor.local_endpoint().port();
    reconnectingSocketStart(port);
    runExpectConnected();
}

TEST_F(ReconnectingSocketTest, connectV6)
{
    acceptorAsyncListenV6();
    auto port = m_acceptor.local_endpoint().port();
    reconnectingSocketStart(port);
    runExpectConnected();
}

TEST_F(ReconnectingSocketTest, reconnect)
{
    acceptorAsyncListenV4();
    auto port = m_acceptor.local_endpoint().port();

    //start reconnectingSocket, it shall connect
    reconnectingSocketStart(port);
    runExpectConnected();

    //reset reconnectingSocket, reconnectingSocket shall reconnect
    m_ioService.post([&]()
    {
        m_reconnectingSocket->reset();
    });
    runExpectReconnected();

    //close server socket, reconnectingSocket shall reconnect
    m_ioService.post([&]()
    {
        m_acceptorSocket.shutdown(boost::asio::ip::tcp::socket::shutdown_both);
        m_acceptorSocket.close();
    });
    runExpectReconnected();
}

TEST_F(ReconnectingSocketTest, connectConnectError)
{
    m_reconnectingSocket->setIdleTimeout(ds::asio::test::LATENCY);
    boost::asio::ip::tcp::endpoint acceptorEndpoint(boost::asio::ip::tcp::v4(), 0);
    m_acceptor.open(acceptorEndpoint.protocol());
    m_acceptor.bind(acceptorEndpoint);
    //no listen

    auto port = m_acceptor.local_endpoint().port();
    reconnectingSocketStart(port);
    ASSERT_EQ(ReconnectingSocket::State::disconnected, m_reconnectingSocket->state());

    //ReconnectingSocket fails to connect and starts idle timer
    ASSERT_RUN_NO_STOP(m_ioService);
    ASSERT_EQ(ReconnectingSocket::State::disconnected, m_reconnectingSocket->state());
    //TODO add stateChanged signal and transition to connecting state

    //make sure we can still connect once server listens
    m_acceptor.listen();
    runExpectConnected();
}

//TODO alternated connections over ipV4 and ipV6

//TODO how to test resolve and connect timeouts? Asio mockups? Reimplement asio platform code?
}} //namespace ds::vdsm
