/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "vdc/VdcFramer.h"

#include <boost/asio/write.hpp>
#include <ds/asio/test.h>

typedef boost::system::error_code error_code;
typedef boost::asio::ip::tcp::resolver resolver;

namespace ds {
namespace vdsm {

class VdcFramerTest : public ::testing::Test
{
public:
    boost::asio::io_service m_ioService;
    VdcFramer m_vdcFramer;
    boost::asio::ip::tcp::acceptor m_acceptor;
    boost::asio::ip::tcp::socket m_socket;
    boost::asio::ip::tcp::socket m_clientSocket;
    VdcFramerTest()
        : m_ioService(1)
        , m_vdcFramer(m_ioService)
        , m_acceptor(m_ioService)
        , m_socket(m_ioService)
        , m_clientSocket(m_ioService)
    {
        boost::asio::ip::tcp::endpoint endpoint(boost::asio::ip::tcp::v4(), 0);
        m_acceptor.open(endpoint.protocol());
        m_acceptor.bind(endpoint);
        m_acceptor.listen();
        m_clientSocket.async_connect(m_acceptor.local_endpoint(), [&](const error_code& e) 
        {
            if (e) {
                FAIL();
            }
        });
        m_acceptor.accept(m_socket);
        m_vdcFramer.setReadTimeout(ds::asio::test::TIMEOUT);
    }
};

TEST_F(VdcFramerTest, write)
{
    m_vdcFramer.start(m_clientSocket);
    m_vdcFramer.asyncWriteFrame("\x40", 1);
    std::vector<char> expected = {'\x00', '\x01', '\x40'};
    std::vector<char> read(3, 0);
    ds::asio::test::runRead(m_socket, boost::asio::buffer(read.data(), read.size()));
    EXPECT_EQ(expected, read);
}

TEST_F(VdcFramerTest, writeBulk)
{
    m_vdcFramer.start(m_clientSocket);
    m_vdcFramer.asyncWriteFrame("\x40", 1);
    m_vdcFramer.asyncWriteFrame("\x40", 1);
    m_vdcFramer.asyncWriteFrame("\x40", 1);
    std::vector<char> expected = {'\x00', '\x01', '\x40'};
    std::vector<char> read(3, 0);
    ds::asio::test::runRead(m_socket, boost::asio::buffer(read.data(), read.size()));
    EXPECT_EQ(expected, read);
    ds::asio::test::runRead(m_socket, boost::asio::buffer(read.data(), read.size()));
    EXPECT_EQ(expected, read);
    ds::asio::test::runRead(m_socket, boost::asio::buffer(read.data(), read.size()));
    EXPECT_EQ(expected, read);
}

TEST_F(VdcFramerTest, read)
{
    bool ok = false;
    m_vdcFramer.start(m_clientSocket);
    m_vdcFramer.setReadFrameFunction([&](const void* data, size_t size)
    {
        ASSERT_EQ(1U, size);
        ASSERT_EQ('\x40', reinterpret_cast<const char*>(data)[0]);
        m_ioService.stop();
        ok = true;
    });
    std::vector<char> writeData = {'\x00', '\x01', '\x40'};
    boost::asio::write(m_socket, boost::asio::buffer(writeData));
    ASSERT_RUN_STOP(m_ioService);
    EXPECT_TRUE(ok);
}

TEST_F(VdcFramerTest, readBulk)
{
    int okCount = 0;
    m_vdcFramer.start(m_clientSocket);
    m_vdcFramer.setReadFrameFunction([&](const void* data, size_t size)
    {
        ASSERT_EQ(1U, size);
        ASSERT_EQ('\x40', reinterpret_cast<const char*>(data)[0]);
        m_ioService.stop();
        okCount ++;
    });
    std::vector<char> writeData = {
        '\x00', '\x01', '\x40',
        '\x00', '\x01', '\x40',
        '\x00', '\x01', '\x40'
    };
    boost::asio::write(m_socket, boost::asio::buffer(writeData));
    ASSERT_RUN_STOP(m_ioService);
    ASSERT_RUN_STOP(m_ioService);
    ASSERT_RUN_STOP(m_ioService);
    EXPECT_EQ(3, okCount);
}

TEST_F(VdcFramerTest, readTimeoutNoActivity)
{
    m_vdcFramer.setReadTimeout(ds::asio::test::LATENCY);
    m_vdcFramer.start(m_clientSocket);
    m_vdcFramer.setReadFrameFunction([&](const void* data, size_t size)
    {
        FAIL(); //not expected to get here
    });
    m_vdcFramer.closed().connect([&]()
    {
        m_ioService.stop();
    });
    ASSERT_RUN_LATENCY_STOP(m_ioService);
}

TEST_F(VdcFramerTest, readTimeoutAfterRead)
{
    m_vdcFramer.setReadTimeout(ds::asio::test::LATENCY);
    m_vdcFramer.start(m_clientSocket);
    int readCount = 0;
    m_vdcFramer.setReadFrameFunction([&](const void* data, size_t size)
    {
        readCount ++;
    });
    m_vdcFramer.closed().connect([&]()
    {
        m_ioService.stop();
    });
    std::vector<char> writeData = {'\x00', '\x01', '\x40'};
    int writeCount = 0;
    //Check that writes restart read timeout
    for (; writeCount < 5; writeCount++) {
        boost::asio::write(m_socket, boost::asio::buffer(writeData));
        ASSERT_RUN_NO_STOP(m_ioService);
    }
    //Check that no write triggers read timeout
    ASSERT_RUN_LATENCY_STOP(m_ioService);
    EXPECT_EQ(writeCount, readCount);
}

TEST_F(VdcFramerTest, closedAfterClose)
{
    m_vdcFramer.start(m_clientSocket);
    bool closedCalled = false;
    m_vdcFramer.closed().connect([&]()
    {
        closedCalled = true;
    });
    m_vdcFramer.close();
    EXPECT_TRUE(closedCalled);
}


TEST_F(VdcFramerTest, closedCalledOnce)
{
    int closedCalledCount = 0;
    m_vdcFramer.closed().connect([&]()
    {
        closedCalledCount ++;
    });
    m_vdcFramer.close();
    EXPECT_EQ(0, closedCalledCount);
    m_vdcFramer.start(m_clientSocket);
    m_vdcFramer.close();
    EXPECT_EQ(1, closedCalledCount);
    m_vdcFramer.close();
    EXPECT_EQ(1, closedCalledCount);
}

}} //namespace ds::vdsm
